/* lex.yy.l - filter lexer
 * Rob Earhart
 * $Id$
 */
/***********************************************************
        Copyright 1997 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

%{
#include <stdlib.h>
#include <string.h>
#include "filter.h"
#include "y.tab.h"

static int lineno = 1;
static int linepos = 0;

static void strnslashcpy(char *to, const char *from, unsigned n);

#define YL (linepos += yyleng)
%}

CRLF (\r\n|\r|\n)

%s COMMENT

%%

#		  { BEGIN(COMMENT);}
<COMMENT>[^\r\n]* { }

[ \t]+		{ YL; }
{CRLF}		{ BEGIN(0); lineno++; linepos=0; }

[0-9]+[GMKgmk]?	{
  char *s;
  yylval.num = strtoul(yytext, &s, 10);
  if (*s)
    switch (*s) {
    case 'G': case 'g':
      yylval.num <<= 10;
    case 'M': case 'm':
      yylval.num <<= 10;
    case 'K': case 'k':
      yylval.num <<= 10;
    }
  YL;
  return NUMBER;
}

\"([^\"]|\\(.|\n))*\" {
  yylval.str = malloc(yyleng - 1);
  strnslashcpy(yylval.str, yytext+1, yyleng - 2);
  yylval.str[yyleng - 2] = '\0';
  YL;
  return STRING;
}

[Ii][Ff]							{ YL; return IF; }
[Tt][Hh][Ee][Nn]						{ YL; return THEN; }
[Ee][Ll][Ss][Ii][Ff]						{ YL; return ELSIF; }
[Ee][Ll][Ss][Ee]						{ YL; return ELSE; }
[Ee][Nn][Dd][Ii][Ff]						{ YL; return ENDIF; }
[Rr][Ee][Qq][Uu][Ii][Rr][Ee]					{ YL; return REQUIRE; }
[Bb][Oo][Uu][Nn][Cc][Ee]					{ YL; return BOUNCE; }
[Ff][Ii][Ll][Ee][Ii][Nn][Tt][Oo]				{ YL; return FILEINTO; }
[Ff][Oo][Rr][Ww][Aa][Rr][Dd]					{ YL; return FORWARD; }
[Nn][Oo][Rr][Mm][Aa][Ll]					{ YL; return NORMAL; }
[Rr][Ee][Pp][Ll][Yy]						{ YL; return REPLY; }
[Ss][Tt][Oo][Pp]						{ YL; return STOP; }
[Tt][Oo][Ss][Ss]						{ YL; return TOSS; }
[Aa][Ll][Ll]-[Oo][Ff]						{ YL; return ALL_OF; }
[Aa][Nn][Yy]-[Oo][Ff]						{ YL; return ANY_OF; }
[Ee][Xx][Ii][Ss][Tt][Ss]					{ YL; return EXISTS; }
[Ff][Aa][Ll][Ss][Ee]						{ YL; return FALSE; }
[Hh][Ee][Aa][Dd][Ee][Rr]					{ YL; return HEADER; }
[Cc][Oo][Nn][Tt][Aa][Ii][Nn][Ss]				{ YL; return CONTAINS; }
[Ii][Ss]							{ YL; return IS; }
[Mm][Aa][Tt][Cc][Hh][Ee][Ss]					{ YL; return MATCHES; }
[Cc][Oo][Nn][Tt][Aa][Ii][Nn][Ss]-[Nn][Oo][Cc][Aa][Ss][Ee]	{ YL; return CONTAINS_NOCASE; }
[Ii][Ss]-[Nn][Oo][Cc][Aa][Ss][Ee]				{ YL; return IS_NOCASE; }
[Mm][Aa][Tt][Cc][Hh][Ee][Ss]-[Nn][Oo][Cc][Aa][Ss][Ee]		{ YL; return MATCHES_NOCASE; }
[Nn][Oo][Tt]							{ YL; return NOT; }
[Ss][Ii][Zz][Ee]						{ YL; return SIZE; }
[Oo][Vv][Ee][Rr]						{ YL; return OVER; }
[Uu][Nn][Dd][Ee][Rr]						{ YL; return UNDER; }
[Ss][Uu][Pp][Pp][Oo][Rr][Tt]					{ YL; return SUPPORT; }
[Tt][Rr][Uu][Ee]						{ YL; return TRUE; }

[a-zA-Z]+	{ YL; return ATOM; }

.               { YL; return yytext[0]; }

%%

int yywrap() {
  return 1;
}

void yyerror(char *s) {
  fprintf(stderr, "%s in line %d, char %d (at `%s')\n",
	  s, lineno, linepos - yyleng, yytext);
}

static void strnslashcpy(char *to, const char *from, unsigned n) {
  for(; n && (*to = *from); n--, to++, from++)
    if (*from == '\\') {
      *to = *++from;
      n++;
    }
}
