/* msg.c - message reader
 * Rob Earhart
 * $Id$
 */
/***********************************************************
        Copyright 1997 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "msg.h"

static const char rcsid[] = "$Id$";

/* Fast&dirty rfc822 message slurper.
 * Does *NOT* implement a lot of the weirder rfc822 things.
 * To be used only for playing around.  XXX */

static MsgHeader headers[100];	/* XXX make this dynamic */
static int nheaders;
static unsigned size;

/* grn grn grn */
#ifndef HAVE_STRDUP
static char *strdup(const char *s) {
  char *ns = malloc(strlen(s) + 1);
  if (ns)
    strcpy(ns, s);
  return ns;
}
#endif /* HAVE_STRDUP */

static int headercmp(const MsgHeader *a,
		     const MsgHeader *b) {
  return strcasecmp(a->name, b->name);
}

int msg_Slurp(FILE *f) {
  char name[80];
  char body[1024];
  unsigned off;
  int c;
  int nheaders = 0;
  enum {
    NAME_START,
    NAME,
    COLON,
    BODY_START,
    BODY
  } state = NAME_START;

  while(1) {
    if ((c = getc(f)) == EOF)
      return -1;
    else {
      switch(state) {
      case NAME_START:
	if (c == '\r' ||
	    c == '\n') {
	  /* we're done! */
	  qsort(headers, nheaders, sizeof(MsgHeader),
		(int (*)(const void *, const void *))&headercmp);
	  headers[nheaders].name = NULL;
	  headers[nheaders].body = NULL;
	  size = 0;
	  if (c == '\r') {
	    int next = getc(f);
	    if (next != '\n' && next != EOF)
	      ungetc(next, f);
	  }
	  while ((c = getc(f)) != EOF)
	    size++;
	  return 0;
	}
	if (! isalpha(c))
	  return -1;
	name[off = 0] = tolower(c);
	state = NAME;
	break;
      case NAME:
	if (c == ' ' ||
	    c == '\t' ||
	    c == ':') {
	  name[++off] = '\0';
	  state = (c == ':' ? BODY_START : COLON);
	  break;
	}
	if (! isalnum(c) &&
	    c != '_')
	  return -1;
	name[++off] = tolower(c);
	break;
      case COLON:
	if (c == ':')
	  state = BODY_START;
	else if (! (c == ' ' ||
		    c == '\t'))
	  return -1;
	break;
      case BODY_START:
	if (c == ' ' ||
	    c == '\t')
	  break;
	if (! isalnum(c))
	  return -1;
	body[off = 0] = c;
	state = BODY;
	break;
      case BODY:
	if (c == '\r' ||
	    c == '\n') {
	  /* Uh oh. */
	  int next = getc(f);
	  if (next == EOF) return -1;
	  if (c == '\r' && next == '\n')
	    next = getc(f);
	  if (next == EOF) return -1;
	  if (next != ' ' && next != '\t') {
	    /* We got one! */
	    body[++off] = '\0';
	    headers[nheaders].name = strdup(name);
	    headers[nheaders].body = strdup(body);
	    nheaders++;
	    state = NAME_START;
	  }
	  ungetc(next, f);
	} else
	  body[++off] = c;
	break;
      }
    }
  }
}

MsgHeader *msg_Header(char *header) {
  MsgHeader *result = bsearch(&header, headers,
			      nheaders, sizeof(MsgHeader),
			      (int (*)(const void *, const void *))&headercmp);
  if (! result)
    result = headers + nheaders;
  else
    while(result > headers
	  && ! strcasecmp(header, result[-1].name))
      result--;

  return result;
}

unsigned msg_Size() {
  return size;
}
