/* y.tab.y - filter parser
 * Rob Earhart
 * $Id$
 */
/***********************************************************
        Copyright 1997 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

%{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "filter.h"

static const char rcsid[] = "$Id$";

extern int filterlex();
static List *Reverse(List *l);
Cmd *result;
%}

%token NUMBER
%token STRING
%token ATOM
%token IF THEN ELSIF ELSE ENDIF REQUIRE
%token BOUNCE FILEINTO FORWARD NORMAL REPLY STOP TOSS
%token ALL_OF ANY_OF EXISTS FALSE HEADER
%token CONTAINS IS MATCHES
%token CONTAINS_NOCASE IS_NOCASE MATCHES_NOCASE
%token NOT SIZE
%token OVER UNDER
%token SUPPORT TRUE

%union {
  unsigned num;
  char *str;
  Cmd *cmd;
  Strings *strs;
  Clause *clause;
  Test *test;
  Tests *tests;
}

%type <str> STRING
%type <num> NUMBER
%type <cmd> commands _commands command action conditional if require
%type <strs> strings _strings
%type <clause> elsifs els
%type <test> test
%type <tests> tests _tests
%type <num> matcher

%%

top: commands {
  result = $1;
}

commands: _commands {
  $$ = (Cmd *)Reverse((List *)$1);
}

_commands: /* empty */ { $$ = NULL; }
         | _commands command {
	   if ($2) $2->list.next = (List *)$1;
	   $$ = $2;
	 }

command: action ';'	{ $$ = $1; }
       | conditional	{ $$ = $1; }

conditional: if | require;

if: IF test THEN commands elsifs els ENDIF {
  If *i;
  Clause *c;
  NEW(i, If);
  NEW(c, Clause);
  i->clause = c;
  c->test = $2;
  c->body = $4;
  if ($5)
    if ($6) {
      /* We have both */
      ((Clause *)$6)->cmd.list.next = (List *)$5;
      c->cmd.list.next = Reverse((List *)$6);
    } else
      /* We only have elsifs */
      c->cmd.list.next = Reverse((List *)$5);
  else
    /* We only have els */
    c->cmd.list.next = (List *)$6;

  $$ = (Cmd *)i;
}

elsifs: /* empty */ { $$ = NULL; }
      | elsifs ELSIF test THEN commands {
	NEW($$, Clause);
	$$->test = $3;
	$$->body = $5;
	$$->cmd.list.next = (List *)$1;
}

els: /* empty */   { $$ = NULL; }
   | ELSE commands {
     NEW($$, Clause);
     $$->test = NULL;
     $$->body = $2;
     $$->cmd.list.next = NULL;
   }

require: REQUIRE STRING {
  Require *r;
  NEW(r, Require);
  r->ext = $2;
  $$ = (Cmd *)r;
}

tests: _tests { $$ = (Tests *)Reverse((List *)$1); }

_tests: test { 
      NEW($$, Tests);
      $$->test = $1;
      $$->list.next = NULL;
    } | _tests ',' test {
      NEW($$, Tests);
      $$->test = $3;
      $$->list.next = (List *)$1;
    }

test: ALL_OF '(' tests ')' {
      All_Of *a;
      NEW(a, All_Of);
      a->tests = $3;
      $$ = (Test *)a;
  } | ANY_OF '(' tests ')' {
      Any_Of *a;
      NEW(a, Any_Of);
      a->tests = $3;
      $$ = (Test *)a;
  } | EXISTS '(' strings ')' {
      Exists *e;
      NEW(e, Exists);
      e->headers = $3;
      $$ = (Test *)e;
  } | FALSE {
      False *f;
      NEW(f, False);
      $$ = (Test *)f;
  } | HEADER '(' strings ')' matcher '(' strings ')' {
      Header *h;
      NEW(h, Header);
      h->headers = $3;
      h->matcher = $5;
      h->keys = $7;
      $$ = (Test *)h;
  } | NOT test {
      Not *n;
      NEW(n, Not);
      n->it = $2;
      $$ = (Test *)n;
  } | SIZE OVER NUMBER {
      Size *s;
      NEW(s, Size);
      s->over = 1;
      s->value = $3;
      $$ = (Test *)s;
  } | SIZE UNDER NUMBER {
      Size *s;
      NEW(s, Size);
      s->over = 0;
      s->value = $3;
      $$ = (Test *)s;
  } | SUPPORT STRING {
      Support *s;
      NEW(s, Support);
      s->extension = $2;
      $$ = (Test *)s;
  } | TRUE {
      True *t;
      NEW(t, True);
      $$ = (Test *)t;
  }

matcher: CONTAINS	 { $$ = CONTAINS; }
       | CONTAINS_NOCASE { $$ = CONTAINS_NOCASE; }
       | IS		 { $$ = IS; }
       | IS_NOCASE	 { $$ = IS_NOCASE; }
       | MATCHES	 { $$ = MATCHES; }
       | MATCHES_NOCASE	 { $$ = MATCHES_NOCASE; }

strings: _strings { $$ = (Strings *)Reverse((List *)$1); }

_strings: STRING {
  		   NEW($$, Strings);
		   $$->str = $1;
		   $$->list.next = NULL;
      } | strings STRING {
	           NEW($$, Strings);
		   $$->str = $2;
		   $$->list.next = (List *)$1;
      }

action: BOUNCE		{
			  Bounce *b;
			  NEW(b, Bounce);
			  $$ = (Cmd *)b;
    } | FILEINTO STRING	{
			  FileInto *f;
			  NEW(f, FileInto);
			  f->folder = $2;
			  $$ = (Cmd *)f;
    } | FORWARD STRING	{
			  Forward *f;
			  NEW(f, Forward);
			  f->where = $2;
			  $$ = (Cmd *)f;
    } | NORMAL		{
			  Normal *n;
			  NEW(n, Normal);
			  $$ = (Cmd *)n;
    } | REPLY STRING	{
			  Reply *r;
			  NEW(r, Reply);
			  r->msg = $2;
			  $$ = (Cmd *)r;
    } | STOP		{
			  Stop *s;
			  NEW(s, Stop);
			  $$ = (Cmd *)s;
    } | TOSS		{
			  Toss *t;
			  NEW(t, Toss);
			  $$ = (Cmd *)t;
    }

%%

static List *Reverse(List *l) {
  List *head = NULL;
  List *it;
  while(l) {
    it = l;
    l = l->next;
    it->next = head;
    head = it;
  }
  return head;
}
